import math
import time
import asyncio
import PyQt5.Qt as qt

nan = float('nan')


def qImage(cvIm):
    """
    Create QImage from cv2 image.
    """
    h, w, ch = cvIm.shape
    qim = qt.QImage(cvIm.data, w, h, w * ch, qt.QImage.Format_RGB888)
    return qim.rgbSwapped()


def run():
    def onTimer():
        loop.call_soon(loop.stop)
        loop.run_forever()

    loop = asyncio.get_event_loop()
    timer = qt.QTimer()
    timer.start(10)
    timer.timeout.connect(onTimer)
    qt.qApp.exec_()


def formatSI(n):
    """
    Format the integer or float n to 3 significant digits + SI prefix.
    """
    s = ''
    if n < 0:
        n = -n
        s += '-'
    if type(n) is int and n < 1000:
        s = str(n) + ' '
    elif n < 1e-22:
        s = '0.00 '
    else:
        assert n < 9.99e26
        log = int(math.floor(math.log10(n)))
        i, j = divmod(log, 3)
        for _try in range(2):
            templ = '%.{}f'.format(2 - j)
            val = templ % (n * 10 ** (-3 * i))
            if val != '1000':
                break
            i += 1
            j = 0
        s += val + ' '
        if i != 0:
            s += 'yzafpnum kMGTPEZY'[i + 8]
    return s


class timeit:
    """
    Context manager for timing.
    """
    def __init__(self, title='Run'):
        self.title = title

    def __enter__(self):
        self.t0 = time.time()

    def __exit__(self, *_args):
        t = time.time() - self.t0
        print(self.title + ' took ' + formatSI(t) + 's')



import asyncio
import threading
import queue


class Runner:
    """
    General processing stage that can be fed input, run some processing
    in a separate thread and make the results available by awaiting
    or as an asynchronous iterator.
    """
    Stop = object()

    def __init__(self):
        self.inQ = queue.Queue()
        self.outQ = asyncio.Queue()
        self.running = False
        self._thread = None
        self._loop = asyncio.get_event_loop()

    def __len__(self):
        return self.outQ.qsize()

    async def __aenter__(self):
        self.start()
        return self

    async def __aexit__(self, *_excinfo):
        self.stop()

    async def __aiter__(self):
        if not self.running:
            self.start()
        return self

    async def __anext__(self):
        if self.running:
            result = await self.outQ.get()
            if result is not Runner.Stop:
                return result
        raise StopAsyncIteration

    def __await__(self):
        return (yield from self.outQ.get())

    def start(self):
        """
        Start processing.
        """
        if self.running:
            return
        self.running = True
        self._thread = threading.Thread(target=self.run)
        self._thread.start()

    def stop(self):
        """
        Stop processing.
        """
        if not self.running:
            return
        self.running = False
        self.inQ.put_nowait(Runner.Stop)
        self.outQ.put_nowait(Runner.Stop)
        self._thread.join()
        self._thread = None
        self.outQ = asyncio.Queue()

    def feed(self, data):
        """
        Feed input to be processed.
        """
        self.inQ.put_nowait(data)
        return self

    __iadd__ = feed

    def output(self, result):
        """
        Add result of processing to the output of this iterator.
        """
        self._loop.call_soon_threadsafe(self.outQ.put_nowait, result)

    def hasOutput(self):
        """
        Is there any output ready?
        """
        return self.outQ.qsize() > 0

    def getInput(self):
        """
        Get next input element.
        """
        return self.inQ.get()

    def getLatestInput(self):
        """
        Get latest input element, skipping over earlier input.
        """
        data = self.inQ.get()
        try:
            while True:
                data = self.inQ.get_nowait()
        except queue.Empty:
            pass
        return data

    def run(self):
        """
        Processing loop. Implementations should stop when self.running
        is False or when Pipeline.Stop is encountered in the input queue.
        """
        pass



import time
import cv2
import string
from collections import namedtuple

from heartwave.runner import Runner


Frame = namedtuple('Frame', ['time', 'image'])


class FrameGrabber(Runner):
    """
    Make a video stream available as a series of frames.

    Input: -
    Output: frame
    """
    def __init__(self, camId=0, width=640, height=480):
        Runner.__init__(self)
        self._args = camId, width, height

    def run(self):
        camId, width, height = self._args
        if type(camId) is int or camId in string.digits:
            camId = int(camId)
            isLive = True
        else:
            isLive = camId.startswith('http://')

        capture = cv2.VideoCapture(camId)
        if capture:
            capture.set(cv2.CAP_PROP_FRAME_WIDTH, width)
            capture.set(cv2.CAP_PROP_FRAME_HEIGHT, height)
            fps = capture.get(cv2.CAP_PROP_FPS)
            while self.running:
                capture.grab()
                t = time.time()
                retval, im = capture.retrieve()
                if not retval:
                    break
                frame = Frame(t, im)
                self.output(frame)
                pause = t - time.time() + 1 / fps
                if not isLive and pause > 0:
                    time.sleep(pause)
            capture.release()
        self.output(Runner.Stop)



import sys
import asyncio
import functools

import PyQt5.Qt as qt

from heartwave.widgets import View, CurveWidget
from heartwave.framegrabber import FrameGrabber
from heartwave.facetracker import FaceTracker
from heartwave.sceneanalyzer import SceneAnalyzer
import heartwave.conf as conf
import heartwave.util as util


class Window(qt.QMainWindow):

    def __init__(self):
        qt.QMainWindow.__init__(self)
        self.setWindowTitle('HeartWave')
        self.view = View(self)
        self.view.setMinimumSize(640, 480)
        self.setCentralWidget(self.view)
        self.curves = CurveWidget()
        self.curves.show()

        def addAction(menu, name, shortcut, cb):
            action = qt.QAction(name, self)
            action.setShortcut(shortcut)
            action.triggered.connect(cb)
            menu.addAction(action)
        menu = self.menuBar()
        fileMenu = menu.addMenu('Source')
        camMenu = fileMenu.addMenu('Camera')
        addAction(fileMenu, 'File', 'Ctrl-F', self.onOpenFile)
        addAction(fileMenu, 'URL', 'Ctrl-U', self.onOpenURL)
        addAction(fileMenu, 'Exit', 'Esc', self.close)
        for i in range(10):
            addAction(camMenu, str(i), '', functools.partial(self.onCamera, i))
        addAction(menu, 'Snapshot', 'Space', self.onSnapshot)
        addAction(menu, 'Toggle curves', 'T', self.onToggleCurves)

        self.pipe = None
        self.running = False
        self.start()

    def onOpenFile(self):
        path, _filter = qt.QFileDialog.getOpenFileName(self)
        self.stop()
        conf.CAM_ID = path
        self.start()

    def onOpenURL(self):
        print('TBD')

    def onCamera(self, camId):
        self.stop()
        conf.CAM_ID = camId
        self.start()

    def onSnapshot(self):
        print('TBD')

    def onToggleCurves(self):
        self.curves.setVisible(not self.curves.isVisible())

    def start(self):
        self.running = True
        self.pipe = asyncio.ensure_future(self.pipeline())

    def stop(self):
        self.running = False
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self.pipe)
        self.pipe = None

    def closeEvent(self, ev):
        self.stop()
        self.curves.close()

    async def pipeline(self):
        """
        Video capture- and processing pipeline that runs asynchronously
        alongside the GUI.
        """
        self.running = True
        async with \
                FrameGrabber(conf.CAM_ID) as grabber, \
                FaceTracker() as tracker, \
                SceneAnalyzer() as analyzer:
            async for frame in grabber:
                if not self.running:
                    break
                tracker.feed(frame)
                if tracker.hasOutput():
                    frame, faces = await tracker
                    analyzer.feed((frame, faces))
                    if analyzer.hasOutput():
                        frame, faces, persons = await analyzer
                        # skip drawing if lagging behind
                        if not grabber.hasOutput():
                            self.view.draw(frame.image, persons)
                            if self.curves.isVisible():
                                self.curves.plot(persons)


def main():
    if len(sys.argv) > 1:
        conf.CAM_ID = sys.argv[1]
    qApp = qt.QApplication(sys.argv)  # noqa
    win = Window()
    win.show()
    util.run()


if __name__ == '__main__':
    main()



from heartwave.person import Person
from heartwave.runner import Runner


class SceneAnalyzer(Runner):
    """
    Processing stage to locate and analyze persons in a scene.

    Input: (frame, faces)
    Output: (frame, faces, persons)
    """
    def run(self):
        persons = []
        while self.running:
            tup = self.getInput()
            if tup is Runner.Stop:
                break
            frame, faces = tup

            present = set()
            for face in faces:
                x, y, w, h = face
                cx = x + w / 2
                cy = y + h / 2
                person = next(
                    (p for p in persons if p.contains(cx, cy)), None)
                if person:
                    person.setFace(face)
                else:
                    person = Person(face)
                    persons.append(person)
                present.add(person)
            persons = [p for p in persons if p in present]

            greenIm = frame.image[:, :, 1]
            for person in persons:
                person.analyze(frame.time, greenIm)

            self.output((frame, faces, persons))
